#!/usr/bin/env python3

"""Collection of request and response header value objects."""

import getpass

from .options import ActionOption, MessageClassOption


class HeaderValue:
    def __bytes__(self):
        raise NotImplementedError


class BytesHeaderValue(HeaderValue):
    """Header with bytes value.

    :param value: Value of the header.
    """

    def __init__(self, value: bytes) -> None:
        self.value = value

    def __str__(self) -> str:
        return f"value={repr(self.value)}"

    def __bytes__(self) -> bytes:
        return self.value

    def __eq__(self, other) -> bool:
        try:
            return self.value == other.value
        except AttributeError:
            return False

    def __repr__(self):
        return f"{self.__class__.__qualname__}(value={repr(self.value)})"


class GenericHeaderValue(HeaderValue):
    """Generic header value."""

    def __init__(self, value: str, encoding="utf8") -> None:
        """Generic header constructor.

        :param value: Value of the header.
        :param encoding: String encoding to use, defaults to "utf8".
        """

        self.value = value
        self.encoding = encoding

    def __str__(self) -> str:
        return f"value={repr(self.value)}, encoding={repr(self.encoding)}"

    def __bytes__(self) -> bytes:
        return self.value.encode(self.encoding)

    def __eq__(self, other) -> bool:
        try:
            return other.value == self.value and other.encoding == self.encoding
        except AttributeError:
            return False

    def __repr__(self):
        return (
            f"{self.__class__.__qualname__}("
            f"value={repr(self.value)}, "
            f"encoding={repr(self.encoding)})"
        )


class CompressValue(HeaderValue):
    """Compress header.  Specifies what encryption scheme to use.  So far only
    'zlib' is supported.
    """

    def __init__(self, algorithm="zlib") -> None:
        """Constructor

        :param algorithm: Compression algorithm to use.  Currently only zlib is supported.
        """

        self.algorithm = algorithm

    def __str__(self) -> str:
        return f"algorithm={repr(self.algorithm)}"

    def __eq__(self, other) -> bool:
        try:
            return self.algorithm == other.algorithm
        except AttributeError:
            return False

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def __bytes__(self) -> bytes:
        return self.algorithm.encode("ascii")


class ContentLengthValue(HeaderValue):
    """ContentLength header.  Indicates the length of the body in bytes."""

    def __init__(self, length: int = 0) -> None:
        """ContentLength constructor.

        :param length: Length of the body.
        """
        self.length = length

    def __int__(self) -> int:
        return self.length

    def __str__(self) -> str:
        return f"length={repr(self.length)}"

    def __eq__(self, other) -> bool:
        try:
            return self.length == other.length
        except AttributeError:
            return False

    def __bytes__(self) -> bytes:
        return str(self.length).encode("ascii")

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(length={repr(self.length)})"


class MessageClassValue(HeaderValue):
    """MessageClass header.  Used to specify whether a message is 'spam' or
    'ham.'
    """

    def __init__(self, value: MessageClassOption = None) -> None:
        """MessageClass constructor.

        :param value: Specifies the classification of the message.
        """

        self.value = value or MessageClassOption.ham

    def __str__(self) -> str:
        return self.value.name

    def __eq__(self, other) -> bool:
        try:
            return self.value == other.value
        except AttributeError:
            return False

    def __bytes__(self) -> bytes:
        return self.value.name.encode("ascii")

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(value={repr(self.value)})"


class SetOrRemoveValue(HeaderValue):
    """Base class for headers that implement "local" and "remote" rules."""

    def __init__(self, action: ActionOption = None) -> None:
        """_SetRemoveBase constructor.

        :param action: Actions to be done on local or remote.
        """

        self.action = action or ActionOption(local=False, remote=False)

    def __str__(self) -> str:
        return f"local={self.action.local}, remote={self.action.remote}"

    def __eq__(self, other) -> bool:
        try:
            return self.action == other.action
        except AttributeError:
            return False

    def __bytes__(self) -> bytes:
        if not self.action.local and not self.action.remote:
            # if nothing is set, then return a blank string so the request
            # doesn't get tainted
            return b""

        values = []
        if self.action.local:
            values.append(b"local")
        if self.action.remote:
            values.append(b"remote")

        return b", ".join(values)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(action={repr(self.action)})"


class SpamValue(HeaderValue):
    """Spam header.  Used by the SPAMD service to report on if the submitted
    message was spam and the score/threshold that it used."""

    def __init__(
        self, value: bool = False, score: float = 0.0, threshold: float = 0.0
    ) -> None:
        """Spam header constructor.

        :param value: True if the message is spam, False if not.
        :param score: Score of the message after being scanned.
        :param threshold: Threshold of which the message would have been marked as spam.
        """

        self.value = value
        self.score = score
        self.threshold = threshold

    def __bool__(self) -> bool:
        return self.value

    def __str__(self) -> str:
        return (
            f"value={str(self.value)}, score={self.score}, threshold={self.threshold}"
        )

    def __eq__(self, other) -> bool:
        try:
            return all(
                [
                    self.value == other.value,
                    self.score == other.score,
                    self.threshold == other.threshold,
                ]
            )
        except AttributeError:
            return False

    def __bytes__(self) -> bytes:
        return b"%b ; %.1f / %.1f" % (
            str(self.value).encode("ascii"),
            self.score,
            self.threshold,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"value={repr(self.value)}, "
            f"score={repr(self.score)}, "
            f"threshold={repr(self.threshold)})"
        )


class UserValue(HeaderValue):
    """User header.  Used to specify which user the SPAMD service should use
    when loading configuration files."""

    def __init__(self, name: str = None) -> None:
        """User constructor.

        :param name: Name of the user account.
        """

        self.name = name or getpass.getuser()

    def __str__(self) -> str:
        return f"name={self.name}"

    def __eq__(self, other) -> bool:
        try:
            return self.name == other.name
        except AttributeError:
            return False

    def __bytes__(self) -> bytes:
        return self.name.encode("ascii")

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(name={repr(self.name)})"



#!/usr/bin/env python3

"""Data structures used for function parameters."""

from enum import Enum
from typing import NamedTuple


class MessageClassOption(Enum):
    """Option to be used for the MessageClass header."""

    spam = "spam"
    ham = "ham"


class ActionOption(NamedTuple):
    """Option to be used in the DidRemove, DidSet, Set, and Remove headers.

    :param local: An action will be performed on the SPAMD service's local database.
    :param remote: An action will be performed on the SPAMD service's remote database.
    """

    local: bool
    remote: bool



#!/usr/bin/env python

"""ConnectionManager classes for TCP and Unix sockets."""

import asyncio
from pathlib import Path
import logging
import ssl
from time import monotonic
from typing import Any, Optional, Tuple

import certifi

from .exceptions import AIOSpamcConnectionFailed, ClientTimeoutException


class Timeout:
    """Container object for defining timeouts."""

    def __init__(
        self, total: float = 600, connection: float = None, response: float = None
    ) -> None:
        """Timeout constructor.

        :param total: The total length of time in seconds to set the timeout.
        :param connection: The length of time in seconds to allow for a connection to live before timing out.
        :param response: The length of time in seconds to allow for a response from the server before timing out.
        """

        self.total = total
        self.connection = connection
        self.response = response

    def __repr__(self):
        return (
            f"{self.__class__.__qualname__}("
            f"total={self.total}, "
            f"connection={self.connection}, "
            f"response={self.response}"
            ")"
        )


class ConnectionManager:
    """Stores connection parameters and creates connections."""

    def __init__(self, timeout: Timeout = None) -> None:
        self.timeout = timeout or Timeout()
        self._logger = logging.getLogger("aiospamc.connections")

    @property
    def logger(self) -> logging.Logger:
        """Return the logger object."""

        return self._logger

    async def request(self, data: bytes) -> bytes:
        """Send bytes data and receive a response.

        :raises: AIOSpamcConnectionFailed
        :raises: ClientTimeoutException

        :param data: Data to send.
        """

        start = monotonic()
        try:
            response = await asyncio.wait_for(self._send(data), self.timeout.total)
        except asyncio.TimeoutError as error:
            self.logger.exception(
                "Timed out (total) to %s after %0.2f seconds",
                self.connection_string,
                monotonic() - start,
                exc_info=error,
                stack_info=True,
                extra={"connection_id": id(self)},
            )
            raise ClientTimeoutException from error

        end = monotonic()
        self.logger.info(
            "Total response time to %s in %0.2f seconds",
            self.connection_string,
            end - start,
            extra={"connection_id": id(self)},
        )

        return response

    async def _send(self, data: bytes) -> bytes:
        reader, writer = await self._connect()

        writer.write(data)
        if writer.can_write_eof():
            writer.write_eof()
        await writer.drain()

        response = await self._receive(reader)

        writer.close()

        return response

    async def _receive(self, reader: asyncio.StreamReader) -> bytes:
        start = monotonic()
        try:
            response = await asyncio.wait_for(reader.read(), self.timeout.response)
        except asyncio.TimeoutError as error:
            self.logger.exception(
                "Timed out receiving data from %s after %0.2f seconds",
                self.connection_string,
                monotonic() - start,
                exc_info=error,
                stack_info=True,
                extra={"connection_id": id(self)},
            )
            raise ClientTimeoutException from error

        end = monotonic()
        self.logger.info(
            "Successfully received data from %s in %0.2f seconds",
            self.connection_string,
            end - start,
            extra={"connection_id": id(self)},
        )

        return response

    async def _connect(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        start = monotonic()
        try:
            reader, writer = await asyncio.wait_for(
                self.open(), self.timeout.connection
            )
        except asyncio.TimeoutError as error:
            self.logger.exception(
                "Timed out connecting to %s after %0.2f seconds",
                self.connection_string,
                monotonic() - start,
                exc_info=error,
                extra={"connection_id": id(self)},
            )
            raise ClientTimeoutException from error

        end = monotonic()
        self.logger.info(
            "Successfully connected to %s in %0.2f seconds",
            self.connection_string,
            end - start,
            extra={"connection_id": id(self)},
        )

        return reader, writer

    async def open(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a connection, returning the reader and writer objects."""

        raise NotImplementedError

    @property
    def connection_string(self) -> str:
        """String representation of the connection."""

        raise NotImplementedError


class TcpConnectionManager(ConnectionManager):
    def __init__(
        self,
        host: str,
        port: int,
        ssl_context: ssl.SSLContext = None,
        timeout: Timeout = None,
    ) -> None:
        """TcpConnectionManager constructor.

        :param host: Hostname or IP address.
        :param port: TCP port.
        :param ssl_context: SSL context.
        """

        super().__init__(timeout)
        self.host = host
        self.port = port
        self.ssl_context = ssl_context

    async def open(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a TCP connection.

        :raises: AIOSpamcConnectionFailed

        :return: Reader and writer for the connection.
        """

        try:
            reader, writer = await asyncio.open_connection(
                self.host, self.port, ssl=self.ssl_context
            )
        except (ConnectionRefusedError, OSError) as error:
            self.logger.exception(
                "Exception occurred when connecting to %s",
                self.connection_string,
                exc_info=error,
                stack_info=True,
                extra={"connection_id": id(self)},
            )
            raise AIOSpamcConnectionFailed from error

        return reader, writer

    @property
    def connection_string(self) -> str:
        """:return: TCP hostname/IP and port."""

        return ":".join([self.host, str(self.port)])


class UnixConnectionManager(ConnectionManager):
    def __init__(self, path: str, timeout: Timeout = None):
        """UnixConnectionManager constructor.

        :param path: Unix socket path.
        """

        super().__init__(timeout)
        self.path = path

    async def open(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a unix socket path connection.

        :raises: AIOSpamcConnectionFailed

        :return: Reader and writer for the connection.
        """

        try:
            reader, writer = await asyncio.open_unix_connection(self.path)
        except (ConnectionRefusedError, OSError) as error:
            self.logger.exception(
                "Exception occurred when connecting to %s",
                self.connection_string,
                exc_info=error,
                stack_info=True,
                extra={"connection_id": id(self)},
            )
            raise AIOSpamcConnectionFailed from error

        return reader, writer

    @property
    def connection_string(self) -> str:
        """:return: Unix connection path."""

        return self.path


def new_ssl_context(verify: Optional[Any]) -> Optional[ssl.SSLContext]:
    """Creates an SSL context based on the supplied parameter.

    :param verify: Use SSL for the connection.  If True, will use root certificates.
        If False, will not verify the certificate.  If a string to a path or a Path
        object, the connection will use the certificates found there.
    """

    if verify is None:
        return None
    elif verify is True:
        return ssl.create_default_context(cafile=certifi.where())
    elif verify is False:
        context = ssl.create_default_context(cafile=certifi.where())
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        return context
    else:
        cert_path = Path(verify).absolute()
        if cert_path.is_dir():
            return ssl.create_default_context(capath=str(cert_path))
        elif cert_path.is_file():
            return ssl.create_default_context(cafile=str(cert_path))
        else:
            raise FileNotFoundError(f"Certificate path does not exist at {verify}")


def new_connection(
    host: Optional[str] = None,
    port: Optional[int] = None,
    socket_path: Optional[str] = None,
    timeout: Optional[Timeout] = None,
    context: Optional[ssl.SSLContext] = None,
) -> ConnectionManager:
    if socket_path:
        return UnixConnectionManager(socket_path, timeout=timeout)
    elif host and port:
        return TcpConnectionManager(host, port, context, timeout)
    else:
        raise ValueError('Either "host" and "port" or "socket_path" must be specified.')



#!/usr/bin/env python3

"""Contains classes used for responses."""

from enum import IntEnum
from typing import Any, Dict, SupportsBytes, Union
import zlib

from .exceptions import *
from .header_values import ContentLengthValue, HeaderValue


class Status(IntEnum):
    EX_OK = 0
    EX_USAGE = 64
    EX_DATAERR = 65
    EX_NOINPUT = 66
    EX_NOUSER = 67
    EX_NOHOST = 68
    EX_UNAVAILABLE = 69
    EX_SOFTWARE = 70
    EX_OSERR = 71
    EX_OSFILE = 72
    EX_CANTCREAT = 73
    EX_IOERR = 74
    EX_TEMPFAIL = 75
    EX_PROTOCOL = 76
    EX_NOPERM = 77
    EX_CONFIG = 78
    EX_TIMEOUT = 79


class Response:
    """Class to encapsulate response."""

    def __init__(
        self,
        version: str = "1.5",
        status_code: Union[Status, int] = 0,
        message: str = "",
        headers: Dict[str, HeaderValue] = None,
        body: bytes = b"",
        **_,
    ):
        """Response constructor.

        :param version: Version reported by the SPAMD service response.
        :param status_code: Success or error code.
        :param message: Message associated with status code.
        :param body: Byte string representation of the body.
        :param headers: Collection of headers to be added.
        """

        self.version = version
        self.headers = headers or {}
        self._status_code: Union[Status, int]
        self.status_code = status_code
        self.message = message
        self.body = body

    def __bytes__(self) -> bytes:
        if "Compress" in self.headers:
            body = zlib.compress(self.body)
        else:
            body = self.body

        if len(body) > 0:
            self.headers["Content-length"] = ContentLengthValue(length=len(body))

        status = self.status_code
        encoded_headers = b"".join(
            [
                b"%b: %b\r\n" % (key.encode("ascii"), bytes(value))
                for key, value in self.headers.items()
            ]
        )
        message = self.message.encode("ascii")

        return (
            b"SPAMD/%(version)b "
            b"%(status)d "
            b"%(message)b\r\n"
            b"%(headers)b\r\n"
            b"%(body)b"
            % {
                b"version": self.version.encode("ascii"),
                b"status": status,
                b"message": message,
                b"headers": encoded_headers,
                b"body": body,
            }
        )

    def __str__(self):
        return (
            f"<{self.status_code} - "
            f"{self.message}: "
            f'{".".join([self.__class__.__module__, self.__class__.__qualname__])} '
            f"object at {id(self)}>"
        )

    def __eq__(self, other: Any) -> bool:
        return (
            self.version == other.version
            and self.headers == other.headers
            and self.status_code == other.status_code
            and self.message == other.message
            and self.body == other.body
        )

    @property
    def status_code(self) -> Union[Status, int]:
        return self._status_code

    @status_code.setter
    def status_code(self, code: Union[Status, int]) -> None:
        try:
            self._status_code = Status(code)
        except ValueError:
            self._status_code = code

    @property
    def body(self) -> bytes:
        return self._body

    @body.setter
    def body(self, value: Union[bytes, SupportsBytes]) -> None:
        self._body = bytes(value)

    def raise_for_status(self) -> None:
        """Raises an exception if the status code isn't zero.

        :raises ResponseException:
        :raises UsageException:
        :raises DataErrorException:
        :raises NoInputException:
        :raises NoUserException:
        :raises NoHostException:
        :raises UnavailableException:
        :raises InternalSoftwareException:
        :raises OSErrorException:
        :raises OSFileException:
        :raises CantCreateException:
        :raises IOErrorException:
        :raises TemporaryFailureException:
        :raises ProtocolException:
        :raises NoPermissionException:
        :raises ConfigException:
        :raises ServerTimeoutException:
        """

        if self.status_code == 0:
            return
        else:
            status_exception = {
                64: UsageException,
                65: DataErrorException,
                66: NoInputException,
                67: NoUserException,
                68: NoHostException,
                69: UnavailableException,
                70: InternalSoftwareException,
                71: OSErrorException,
                72: OSFileException,
                73: CantCreateException,
                74: IOErrorException,
                75: TemporaryFailureException,
                76: ProtocolException,
                77: NoPermissionException,
                78: ConfigException,
                79: ServerTimeoutException,
            }
            if self.status_code in status_exception:
                raise status_exception[self.status_code](self.message)
            else:
                raise ResponseException(self.status_code, self.message)



#!/usr/bin/end python

"""Frontend functions for the package."""

import logging
import time
from typing import (
    Any,
    Callable,
    Dict,
    NamedTuple,
    Optional,
    SupportsBytes,
    Type,
    Union,
)
from ssl import SSLContext

from .connections import Timeout, new_connection, new_ssl_context, ConnectionManager
from .exceptions import BadResponse, ParseError
from .header_values import CompressValue, MessageClassValue, UserValue
from .options import ActionOption, MessageClassOption
from .incremental_parser import ResponseParser, parse_set_remove_value
from .responses import Response
from .requests import Request


ConnectionFactory = Callable[
    [
        Optional[str],
        Optional[int],
        Optional[str],
        Optional[Timeout],
        Optional[SSLContext],
    ],
    ConnectionManager,
]
SSLFactory = Callable[[Any], Optional[SSLContext]]


class Client(NamedTuple):
    ssl_context_factory: SSLFactory
    connection_factory: ConnectionFactory
    parser_factory: Type[ResponseParser]


DEFAULT_CLIENT = Client(new_ssl_context, new_connection, ResponseParser)

LOGGER = logging.getLogger("aiospamc")


async def request(
    req: Request, connection: ConnectionManager, parser: ResponseParser
) -> Response:
    """Sends a request and returns the parsed response.

    :param req: The request to send.
    :param connection: An instance of a connection.
    :param parser: An instance of a response parser.

    :return: The parsed response.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    """

    start = time.monotonic()
    LOGGER.info(
        "Sending %s request",
        req.verb,
        extra={
            "message_id": id(req.body),
            "request_id": id(req),
            "connection_id": id(connection),
        },
    )
    response = await connection.request(bytes(req))
    try:
        parsed_response = parser.parse(response)
    except ParseError as error:
        LOGGER.exception(
            "Error parsing response",
            exc_info=error,
            extra={"request_id": id(req), "connection_id": id(connection)},
        )
        raise BadResponse(response) from error
    response_obj = Response(**parsed_response)
    response_obj.raise_for_status()
    end = time.monotonic()
    LOGGER.info(
        "Successfully received response in %0.2f",
        end - start,
        extra={"request_id": id(req), "connection_id": id(connection)},
    )

    return response_obj


def _new_connection(
    connection_factory: ConnectionFactory,
    ssl_context_factory: SSLFactory,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Any = None,
) -> ConnectionManager:
    ssl_context = ssl_context_factory(verify)
    connection = connection_factory(host, port, socket_path, timeout, ssl_context)

    return connection


def _add_headers(req: Request, user: Optional[str], compress: Optional[bool]) -> None:
    if user:
        req.headers["User"] = UserValue(user)
    if compress:
        req.headers["Compress"] = CompressValue()


async def check(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("CHECK", body=bytes(message))
    _add_headers(req, user, compress)
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def headers(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return the modified message headers.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains the modified
        message headers, but not the content of the message.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("HEADERS", body=bytes(message))
    _add_headers(req, user, compress)
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def ping(
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    **kwargs,
) -> Response:
    """Sends a ping to the SPAMD service.

    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.

    :return: A response with "PONG".

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("PING")
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def process(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a modified
        copy of the message.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("PROCESS", body=bytes(message))
    _add_headers(req, user, compress)
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def report(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a report.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("REPORT", body=bytes(message))
    _add_headers(req, user, compress)
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def report_if_spam(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a report if
        the message is considered spam.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("REPORT_IFSPAM", body=bytes(message))
    _add_headers(req, user, compress)
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def symbols(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return a response with rules that matched.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a
        comma-separated list of the symbols that were hit.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    req = Request("SYMBOLS", body=bytes(message))
    _add_headers(req, user, compress)
    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)


async def tell(
    message: Union[bytes, SupportsBytes],
    message_class: Union[str, MessageClassOption],
    remove_action: Union[str, ActionOption] = None,
    set_action: Union[str, ActionOption] = None,
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: str = None,
    timeout: Timeout = None,
    verify: Optional[Any] = None,
    user: str = None,
    compress: bool = False,
    **kwargs,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param message_class: Classify the message as 'spam' or 'ham'.
    :param remove_action: Remove message class for message in database.
    :param set_action: Set message class for message in database.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :module:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with "DidSet" and/or "DidRemove" headers along with the
        actions that were taken.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    client = kwargs.get("client", DEFAULT_CLIENT)

    headers: Dict[str, Any] = {
        "Message-class": MessageClassValue(MessageClassOption(message_class))
    }
    if remove_action:
        headers["Remove"] = parse_set_remove_value(remove_action)
    if set_action:
        headers["Set"] = parse_set_remove_value(set_action)
    req = Request("TELL", headers=headers, body=bytes(message))
    _add_headers(req, user, compress)

    connection = _new_connection(
        client.connection_factory,
        client.ssl_context_factory,
        host,
        port,
        socket_path,
        timeout,
        verify,
    )
    parser = client.parser_factory()

    return await request(req, connection, parser)



#!/usr/bin/env python3

"""Collection of exceptions."""


class ClientException(Exception):
    """Base class for exceptions raised from the client."""

    pass


class BadRequest(ClientException):
    """Request is not in the expected format."""

    pass


class BadResponse(ClientException):
    """Response is not in the expected format."""

    pass


class AIOSpamcConnectionFailed(ClientException):
    """Connection failed."""

    pass


class ResponseException(Exception):
    """Base class for exceptions raised from a response."""

    def __init__(self, code: int, message: str) -> None:
        self.code = code
        super().__init__(message)


class UsageException(ResponseException):
    """Command line usage error."""

    def __init__(self, message: str) -> None:
        super().__init__(64, message)


class DataErrorException(ResponseException):
    """Data format error."""

    def __init__(self, message: str) -> None:
        super().__init__(65, message)


class NoInputException(ResponseException):
    """Cannot open input."""

    def __init__(self, message: str) -> None:
        super().__init__(66, message)


class NoUserException(ResponseException):
    """Addressee unknown."""

    def __init__(self, message: str) -> None:
        super().__init__(67, message)


class NoHostException(ResponseException):
    """Hostname unknown."""

    def __init__(self, message: str) -> None:
        super().__init__(68, message)


class UnavailableException(ResponseException):
    """Service unavailable."""

    def __init__(self, message: str) -> None:
        super().__init__(69, message)


class InternalSoftwareException(ResponseException):
    """Internal software error."""

    def __init__(self, message: str) -> None:
        super().__init__(70, message)


class OSErrorException(ResponseException):
    """System error (e.g. can't fork the process)."""

    def __init__(self, message: str) -> None:
        super().__init__(71, message)


class OSFileException(ResponseException):
    """Critical operating system file missing."""

    def __init__(self, message: str) -> None:
        super().__init__(72, message)


class CantCreateException(ResponseException):
    """Can't create (user) output file."""

    def __init__(self, message: str) -> None:
        super().__init__(73, message)


class IOErrorException(ResponseException):
    """Input/output error."""

    def __init__(self, message: str) -> None:
        super().__init__(74, message)


class TemporaryFailureException(ResponseException):
    """Temporary failure, user is invited to try again."""

    def __init__(self, message: str) -> None:
        super().__init__(75, message)


class ProtocolException(ResponseException):
    """Remote error in protocol."""

    def __init__(self, message: str) -> None:
        super().__init__(76, message)


class NoPermissionException(ResponseException):
    """Permission denied."""

    def __init__(self, message: str) -> None:
        super().__init__(77, message)


class ConfigException(ResponseException):
    """Configuration error."""

    def __init__(self, message: str) -> None:
        super().__init__(78, message)


class TimeoutException(Exception):
    """General timeout exception."""

    pass


class ServerTimeoutException(ResponseException, TimeoutException):
    """Timeout exception from the server."""

    def __init__(self, message: str) -> None:
        super().__init__(79, message)


class ClientTimeoutException(ClientException, TimeoutException):
    """Timeout exception from the client."""

    pass


class ParseError(Exception):
    """Error occurred while parsing."""

    def __init__(self, message=None) -> None:
        """Construct parsing exception with optional message.

        :param message: User friendly message.
        """

        self.message = message


class NotEnoughDataError(ParseError):
    """Expected more data than what the protocol content specified."""

    pass


class TooMuchDataError(ParseError):
    """Too much data was received than what the protocol content specified."""

    pass
